{{
@file pjvm_demo.spin:
  A no-debug demo PASM version of Peter's Spin JVM.
}}

con                                                              
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  '_clkmode = xtal1 + pll8x
  '_xinfreq = 10_000_000

obj
  'sx : "FullDuplexSingleton"
  sx : "SingleSerial"
  pm : "pjmemory"               ' best fit list power of 2 size
  pj : "pjvmInterpreter"
  pjd: "pjdata"
  pcn: "pjcon"
  pn : "pnative"
    
pub main | progptr, ctrlp, cmdptr
{{
Startup the JVM interpreter
}}
  'sx.start(31,30,0,115200)     ' full duplex
  sx.init(31,30,19200)          ' SingleSerial
  waitcnt(clkfreq/4+cnt)
  'sx.str(string($d,"JVM Startup"))
    
  ctrlp := pjd.getControlPtr
  cmdptr := pjd.getCmdPtr

  pj.start(cmdptr)              ' start jvm engine
  
  progptr := pj.getProgPtr
  longmove(pj.getProgBuffer, progptr, pj.getProgSize/4)
  progptr := pj.getProgBuffer
  
  pjd.init(progptr,pj.getJavaCodeSize,1,false)
  pm.init(pjd.getHS-2)

  pn.start(3) ' start 3*2 = 6 virtual peripherals

  
'{ add tick (') comment this to enable pasm runner. uncomment to use spin runner
  pasmrun(progptr, cmdptr, ctrlp)
  repeat ' should never get here

'var long stack[$100]

dat
  ' program control variables
  mypc        long 0    ' storage for PC pointer        ... keep in order
  myprog      long 0    ' storage for PROG pointer      ... keep in order
  mycmd       long 0    ' storage for CMD pointer       ... keep in order
  myctrl      long 0    ' storage for CTRL pointer      ... keep in order
  mywait      long $55  ' storage for Spin WAIT pointer ... keep in order

  ' garbage collection variables
  gprogstart  long 0    ' program start location 
  gstatictop  long 0    ' static end location - heapbase
  gheaptop    long 0    ' heap start location - also static start
  gstackbase  long 0    ' stack base location
  gheapptr    long 0    ' dynamic
  gstkptr     long 0    ' dynamic

dat
'______________________________________________________________
'______________________________________________________________
' the runner
runner        org       0
'______________________________________________________________
' initialize
'
long 0 [8]    'nops for debugger
              mov       _temp,    par
              mov       _pcptr,   _temp         ' pointer to PC
              add       _temp, #4
              mov       _progptr, _temp         ' pointer to PROGRAM
              add       _temp, #4
              mov       _cmdptr,  _temp         ' pointer to CMD
              add       _temp, #4
              mov       _ctrlptr, _temp         ' pointer to CONTROL
              add       _temp, #4            
              mov       _waitptr, _temp         ' pointer to SPIN WAIT

'----------------------------------
' GC support (garbage collection)
'
              add       _temp, #4            
              mov       _progstart,_temp
              rdlong    _progstart,_progstart   ' always same
              add       _temp, #4            
              mov       _statictop,_temp         
              rdlong    _statictop,_statictop   ' always same
              add       _temp, #4            
              mov       _heaptop, _temp
              rdlong    _heaptop, _heaptop      ' always same
              mov       _staticend,_heaptop
              sub       _staticend,#2           ' sub to simplify checking last static element
              add       _heaptop,#2             ' add to simplify checking last heap element
              add       _temp, #4            
              mov       _stackbase,_temp         
              rdlong    _stackbase,_stackbase   ' always same
              sub       _stackbase,#2           ' sub to simplify checking last stack element
              
              add       _temp, #4            
              mov       _heapptr, _temp         ' dynamic    
              rdword    _heapptr, _temp
              add       _temp, #4            
              mov       _stkptr,  _temp         ' dynamic
              rdword    _stkptr,  _temp
'
' end GC support variables
'----------------------------------

'______________________________________________________________
' adjust pointers
'
              rdlong    _pcptr,   _pcptr     ' pointer to PC
              rdlong    _cmdptr,  _cmdptr    ' pointer to CMD
              rdlong    _ctrlptr, _ctrlptr   ' pointer to CTRL

              wrlong    _zero,    _waitptr   ' let spin starter know we're ready

              mov       _gccount, _gcepoch

'______________________________________________________________
' process bytecode instructions
'
cmdloop              
' get bytecode
' cmd := getCmd(progptr+pc)
              rdlong    _mypc,  _pcptr
              rdlong    _temp,  _progptr
              add       _mypc,  _temp

' get cmd bytecode from pc
              rdbyte    _mcmd,  _mypc

'{ do garbage collection before NEW, NEWARRAY, and ANEWARRAY
              cmp       _mcmd,  #$57 wz
        if_ne cmp       _mcmd,  #$59 wz
        if_ne cmp       _mcmd,  #$5b wz
        if_e  call      #GC
'}

' long[jmptr] := cmd
              wrlong    _mcmd,  _cmdptr

' set long[ctrlp] := non-zero
              wrlong    _minus1,_ctrlptr

' wait while long[ctrlp] is non-zero
waitresp      rdlong    _resp,  _ctrlptr wz
        if_nz jmp       #waitresp
              rdlong    _resp,  _cmdptr

              cmps      _resp,  #$100 wc
        if_c  jmp       #cmdloop 'if d < s
                                     
'______________________________________________________________
' native and exception handling
'

' if long[jmptr] & $8000, throw exception
' else if long[jmptr] & $100, handle native

              cmp       _resp,   #$100  wc      ' if resp > $ff check for native IO
        if_c  jmp       #cmdloop

              test      _resp,  _x8100  wz      ' is native or exception?
        if_z  jmp       #cmdloop                ' not native or exception ... next instruction

' native or exception
special       wrlong    _resp,  _waitptr        ' tell spin to do work
              rdlong    _resp,  _waitptr wz     ' exceptions will hang by design    
              tjnz      _resp,  #$-1            ' wait for spin native to finish
        if_c  wrlong    _minus1,_ctrlptr
        if_c  jmp       #waitresp               ' special
        
              test      _resp,  _x8000  wc      ' if exception stall
        if_c  jmp       #$                      ' stall

              jmp       #cmdloop                ' next

'______________________________________________________________
' run through stack, static, and heap to find unused references
'
_gccount      long      0       ' GC skips if count <> 0
_gcepoch      long      31 '7       ' epoch = 8 is a fair optimization
' fibo23 in a little less than 4 seconds
' Object counts: fibo14=31 fibo15=32 fibo28=32

' Strangely, if we change the epoch, println behaves erratically.


GC  ' garbage collector

' time to collect garbage?
              djnz      _gccount, #GC_ret       ' if gccount <> 0, skip collection
              mov       _gccount, _gcepoch

' walk through heap ... heapptr is dynamic so we save a pointer to the pointer and read it
              rdword    _nextptr,  _heapptr     ' get current heap pointer
gc_next       mov       _prevptr,  _nextptr     ' save pointer for free
              rdword    _nextptr,  _nextptr     ' get next heap pointer
              mov       _temp,     _nextptr     ' get ready to read next pointer
' if node is used LSB is set ... rdword ignores LSB so not important here
              rdword    _temp,     _temp        ' get next pointer

' run through loop looking for _curptr in stack, static, heap tables
'              
gc_loop       tjz       _temp,     #GC_ret      ' done if nextptr is 0  

' get data pointer
              mov       _currptr,  _nextptr     ' get the link
              andn      _currptr,  #1           ' ignore free bit
              add       _currptr,  #2           ' read data pointer

' check _currptr against static
              mov       _temp,     _statictop   ' start at top and go down
statloop      rdword    mresp,     _temp        ' get word
              cmp       mresp,     _currptr wz  ' match?
        if_e  jmp       #gc_gotit               ' yes match, don't free
              sub       _temp,     #2           ' next location
              cmp       _temp,     _staticend wz' if not finished, get next data
        if_ne jmp       #statloop               ' else do next section

' check _currptr against stack
              rdword    _temp,     _stkptr      ' start at top and go down
              add       _temp,     _progstart   ' add program offset for stack comparisons
stkloop       rdword    mresp,     _temp        ' get word
              call      #swap                   ' objects stored swapped in stack
              cmp       mresp,     _currptr wz  ' match?
        if_e  jmp       #gc_gotit               ' yes match, don't free
              sub       _temp,     #2           ' next location
              cmp       _temp,     _stackbase wz' if not finished, get next data
        if_ne jmp       #stkloop                ' else do next section

' check _currptr against heap
              rdword    _temp,     _heapptr     ' get heaphead and go bottom up
hploop        rdword    mresp,     _temp        ' get word from heap
              call      #swap                   ' objects stored swapped in heap
              cmp       mresp,     _currptr wz  ' heap word and _currptr match?
        if_e  jmp       #gc_gotit               ' yes match, don't free
              add       _temp,     #2           ' next location
              cmp       _temp,     _heaptop wz  ' if not finished, get next data
        if_ne jmp       #hploop                 ' else do next section

' if we get to here, the pointer is not in use, set pointer free
              andn      _nextptr,  #1           ' clear bit to free it              
              wrword    _nextptr,  _prevptr
gc_gotit      jmp       #gc_next

'gc_gotit      jmp       #gc_next               ' easier to debug if gotit on match
                                                       
GC_ret        ret

'______________________________________________________________
' swap bytes between lanes
'
swapvar       long 0
swap          mov       swapvar,    mresp
              shr       swapvar,    #8
              shl       mresp,      #8
              or        mresp,      swapvar
              and       mresp,      _xffff
swap_ret      ret

'______________________________________________________________
'
_minus1       long      -1
_x8100        long      $8100
_x8000        long      $8000
_xffff        long      $ffff
_malloc       long      pj#MALLOC_BIT '| pj#MALLOC_BIT << 1
_zero         long      0

_pcptr        res 1
_progptr      res 1
_cmdptr       res 1
_ctrlptr      res 1
_waitptr      res 1
_mypc         res 1
_mcmd         res 1
_temp         res 1
_resp         res 1
mresp         res 1

' need these for garbage collection
'
_progstart    res 1     
_heaptop      res 1     ' also static start
_statictop    res 1
_staticend    res 1
_stackbase    res 1
_heapptr      res 1     ' dynamic
_stkptr       res 1     ' dynamic

_prevptr      res 1     ' prev pointer contains node to free if any
_nextptr      res 1     ' next in heap
_currptr      res 1     ' current search heap pointer

'______________________________________________________________

pub pasmrun(progp, cmdptr, ctrlp) | mycog

'  sx.str(string($d,"Start PASMRUN ... "))

  mypc        := pjd.getPCptr
  myprog      := progp
  mycmd       := cmdptr
  myctrl      := ctrlp
  gprogstart  := progp
  gstatictop  := pjd.getHB
  gheaptop    := pjd.getHS
  gstackbase  := pjd.getSB+gprogstart
  gheapptr    := pjd.getRefHP
  gstkptr     := pjd.getRefSP  

{
  sx.str(string($d,"statictop "))
  sx.hex(gstatictop,4)
  sx.str(string($d,"heaptop   "))
  sx.hex(gheaptop,4)
  sx.str(string($d,"heapptr   "))
  sx.hex(gheapptr,4)
  sx.str(string($d,"heaphead  "))
  sx.hex(word[gheapptr],4)
  sx.str(string($d,"stackbase "))
  sx.hex(gstackbase,4)
  sx.str(string($d,"stackptr  "))
  sx.hex(gstkptr,4)
  sx.str(string($d,"stacktop  "))
  sx.hex(word[gstkptr]+gprogstart,4)
  sx.out($d)
'}

'  cognew(handler, @hstack)
'  bu.taskstart(@runner, @mypc, string("RUNNER"))
'  bu.start ' does not return

  'mycog  := pr.start(@mypc)
  mycog := cognew(@runner, @mypc)+1
  repeat while mywait           ' wait for COG to start

  handler

'obj bu : "BmaUtility"
'var long hstack[$40]
'obj pr : "pjrunner"


pub handler | cmd, tmp
{{
handle memory allocations, natives, and exceptions
}}       
  repeat

    if (long[myctrl] > 0) and (long[myctrl] & pj#MALLOC_BIT)
      tmp := long[myctrl] & $3fff
      pm.alloc(tmp)
      'pm.simplealloc(tmp)
      long[myctrl] := -1

    if mywait & pj#EXCPTN_BIT

      'cmd := getCmd(pjd.getProgPtr+pjd.getPC)
      sx.str(string($d,"Exception "))
      sx.hex(mywait,4)
      sx.str(string(" at "))
      sx.hex(pjd.getPC,4)
      sx.tx($d)
      mywait := 0
      repeat ' stall forever

    elseif mywait & $100

      cmd := mywait & $3f
      pn.doNative(cmd)
      pjd.incrPC(3)
      mywait := 0

'______________________________________________________________
'}
' end of main if pasmrunner code commented

{ add tick (') comment this and uncomment pasmrun to use spin runner
  run(progptr,cmdptr,ctrlp)
  repeat
              
'var
'  long  acnt

pub run(progptr, cmdptr, ctrlp) | cmd, deb, mpc, opc, trace, n, m, op1, op2, brk

  deb  := -1  ' run
'  acnt := 0
  trace~
'  trace~~  
  
  mpc := pjd.getPC
  cmd := getCmd(progptr+mpc) 'byte[progptr+mpc]
  opc := mpc
  op1 := byte[progptr+mpc+1]
  op2 := byte[progptr+mpc+2]

  long[cmdptr] := cmd

  repeat

    long[ctrlp]~~              ' step
    repeat 
      if (long[ctrlp] > 0) and (long[ctrlp] & pj#MALLOC_BIT)
        pm.alloc(long[ctrlp] & $3fff)
        'pm.simplealloc(long[ctrlp] & $3fff)
        long[ctrlp] := -1
    while long[ctrlp] ' wait for command to be completely finished regardless of MALLOC_BIT
      
    mpc := pjd.getPC
    cmd := getCmd(progptr+mpc) 'byte[progptr+mpc]

    'if mpc == $92F
    '  trace~~

    if mpc <> opc
       
'{  
      'if opc == $489 or opc == $487
      'if opc > $470 and opc < $490
      'if opc > $500 and opc < $600
      'if trace
      {
        printf(string($d,"PC $%4x $%4x: $%02x $%02x $%02x %13s H%04x M%04x F%04x S%04x "),{
        }p(opc)+p(pjd.getPC)+p(byte[cmdptr])+p(op1)+p(op2)+p(pcn.opcodeNamePtr(long[cmdptr])){
        }+p(pjd.getHP)+p(pjd.getMP)+p(pjd.getFP+progptr)+p(pjd.getSP+progptr))
        dumpWipFrame(progptr)
      }
      printf(string($d,"PC $%4x $%4x: $%02x $%02x $%02x %13s H%04x C%d M%04x F%04x S%04x "),{
      }p(opc)+p(pjd.getPC)+p(byte[cmdptr])+p(op1)+p(op2)+p(pcn.opcodeNamePtr(long[cmdptr])){
      }+p(pjd.getHP)+p(pjd.isCarry)+p(pjd.getMP)+p(pjd.getFP+progptr)+p(pjd.getSP+progptr))
      'dumpWipFrame(progptr)
      dumpWipLastStack(progptr)

{
      if(byte[cmdptr] == pcn#Jem_IALOAD)
        dumpStack(progptr)
        dumpStatic
        dumpHeap
      if(byte[cmdptr] == pcn#Jem_IASTORE)
        dumpStack(progptr)
        dumpStatic
        dumpHeap
}
'}
      opc := mpc
      'acnt++

    if long[cmdptr] > $ff

      'if long[cmdptr] & pj#MALLOC_BIT
      
      if long[cmdptr] & $8000
        sx.str(string($d,"Exception "))
        sx.hex(long[cmdptr],4)
'{
        sx.str(string(" Exception in instruction."))
        printf(string($d," PC $%4x: $%02x $%02x $%02x %13s M%04x H%04x F%04x S%04x "),{
        }p(opc)+p(cmd)+p(op1)+p(op2)+p(pcn.opcodeNamePtr(cmd)){
        }+p(pjd.getMP)+p(pjd.getHP)+p(pjd.getFP)+p(pjd.getSP))
        dumpWipFrame(progptr)
        printf(string($d),0)

        dumpStack(progptr)
        dumpStatic
        dumpHeap
        pm.show
        'deb := getch
'}
        repeat                  ' hang on exception
                
      elseif long[cmdptr] & $100
'{      
        printf(string($d,"PC $%4x $%4x: $%02x $%02x $%02x N%12s H%04x C%d M%04x F%04x S%04x "),{
        }p(opc)+p(pjd.getPC)+p(cmd)+p(op1)+p(op2)+p(pcn.opcodeNamePtr(cmd)){
        }+p(pjd.getHP)+p(pjd.isCarry)+p(pjd.getMP)+p(pjd.getFP+progptr)+p(pjd.getSP+progptr))
        'dumpWipFrame(progptr)
        dumpWipLastStack(progptr)
'}
        cmd :=  long[cmdptr] & $3f

        pn.doNative(cmd)
        pjd.incrPC(3)

    long[cmdptr] := getCmd(progptr+pjd.getPC)
    op1 := byte[progptr+mpc+1]
    op2 := byte[progptr+mpc+2]


pri getCmd(ptr)
  result := byte[ptr]
  ifnot result ' engine can't handle nop = 0
    result := byte[ptr+1]

obj
  pf : "snprintf"

dat
  pb byte 0 [256]
  
pri printf(fmt, arg)
  pf.snprintf(@pb,255,fmt,arg)
  sx.str(@pb)

pri p(v)
  return pf.p(v)


pub dumpWipFrame(progptr) | n
  repeat n from 0 to pjd.getSP-pjd.getFP-2 step 2
    printf(string("$%04x "),p(swp16(word[progptr+n+pjd.getFP])))

pub dumpWipLastStack(progptr) | n
  repeat n from pjd.getSP-pjd.getFP-8 to pjd.getSP-pjd.getFP-2 step 2
    printf(string("$%04x "),p(swp16(word[progptr+n+pjd.getFP])))

pub dumpStack(progptr) | n, mpc
  printf(string($d,"Stack: "),0)
  mpc := pjd.getSP-pjd.getBP
  repeat n from 0 to 30 step 2
    printf(string("$%04x "),p(n+pjd.getBP))
  repeat n from 0 to mpc-2 step 2
    if n // 32 == 0
      printf(string($d,"$%04x: "),p(n+pjd.getBP))
    printf(string("$%04x "),p(swp16(word[progptr+n+pjd.getBP])))
'  printf(string($d),0)


pub dumpStatic | n, hs,stap
  stap := pjd.getStaticPtr
  hs := pjd.getHB - stap
  repeat n from 0 to hs step 2
    if (n // 32) == 0
      printf(string($d,"s%04x "),p(stap+n))
    printf(string("%04x "),p(word[stap+n]))

pub dumpHeap | n, hs
  hs := pjd.getHS-pjd.getHP - 2
  if hs < 2
     return
  printf(string($d,"h%04x "),p(pjd.getHP))
  repeat n from 0 to hs step 2
    if ((pjd.getHP+n) // 32) == 0
      printf(string($d,"h%04x "),p(pjd.getHP+n))
    printf(string("%04x "),p(swp16(word[pjd.getHP+n])))

pub swp16(val)
  result := val & $ff
  result <<=8
  val >>=8
  result |= val
'}  

      